#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* Misc utilities */

typedef enum { false, true } Flag;

void
error (const char *format, ...)
{
  va_list args;
  va_start (args, format);
  vfprintf (stderr, format, args);
  va_end (args);
  fprintf (stderr, "\n");
  exit (1);
}


void
stdlib_error (const char *context)
{
  error ("%s: %s", context, strerror (errno));
}


void *
allot (size_t size)
{
  void *result = malloc (size);
  if (!result && size != 0)
    stdlib_error("Out of memory");
  return result;
}


FILE *
open_file (const char *filename, const char *mode)
{
  if (strcmp (filename, "-") == 0)
    {
      assert (mode [0] == 'r' || mode [0] == 'w' || mode [0] == 'a');
      return mode [0] == 'r' ? stdin : stdout;
    }
  else
    {
      FILE *result = fopen (filename, mode);
      if (!result)
	stdlib_error (filename);
      return result;
    }
}


/* Elements */

typedef enum { R, G, B_color, C, M, Y, K, W, Root_color } Color;

enum { Root_size = 10 };

typedef struct
{
  unsigned int b: 1;
  unsigned int em: 1;
  unsigned int i: 1;
  unsigned int s: 1;
  unsigned int tt: 1;
  unsigned int u: 2;
  unsigned int size: 4;
  unsigned int color: 4;
} 
Context;

typedef struct 
{
  Context context;
  char character;
}
Element;


static Flag
equal_contexts (Context c1, Context c2)
{
  return c1.b == c2.b 
    && c1.em == c2.em
    && c1.i == c2.i
    && c1.s == c2.s
    & c1.tt == c2.tt
    && c1.u == c2.u
    && c1.size == c2.size
    && c1.color == c2.color;
}


static void
check_context (Context context)
{
  assert (context.size <= Root_size);
  assert (R <= context.color && context.color <= Root_color);
}


static void
check_element (Element e)
{
  check_context (e.context);
  assert (isprint (e.character) || 
	  e.character == '\t' ||
	  e.character == '\r' ||
	  e.character == '\n');
}


static Context
make_root_context (void)
{
  Context result;
  result.b = result.em = result.i = result.s = result.tt = result.u = 0;
  result.size = Root_size;
  result.color = Root_color;
  check_context (result);
  return result;
}


/* Buffers */

typedef struct 
{
  int size, ptr;
  Element *elements;
}
Buffer;


static Buffer *
make_buffer (void)
{
  Buffer *b = allot (sizeof *b);
  b->size = 5 * 1024 * 1024;	/* note 5 mega-character limit */
  b->ptr = 0;
  b->elements = allot (b->size * sizeof b->elements[0]);
  return b;
}


static void 
append (Buffer *buffer, Context context, char ch)
{
  check_context (context);
  if (buffer->size <= buffer->ptr )
    error ("Input too big");
  buffer->elements[buffer->ptr].context = context;
  buffer->elements[buffer->ptr].character = ch;
  buffer->ptr++;
}


/* Return true iff the previous output character in buffer was a SPC
   decorated with the same space context as sc. */
static Flag
was_whitespace (const Buffer *buffer, Context sc)
{
  check_context (sc);
  if (0 == buffer->ptr)
    return false;
  else
    {
      const Element *e = buffer->elements + buffer->ptr - 1;
      return e->character == ' ' && equal_contexts (e->context, sc);
    }
}


static Flag
equal (const Buffer *b1, const Buffer *b2)
{
  if (b1->ptr != b2->ptr)
    return 0;
  return
    0 == memcmp (b1->elements, b2->elements, b1->ptr * sizeof b1->elements[0]);
}


/* Canonicalization */

static Context
space_context (Context c)
{
  Context r = c;
  r.s = r.em = r.i = r.b = 0;
  if (r.u == 0)
    r.color = W;
  check_context (r);
  return r;
}


static void
append_whitespace_char (Buffer *buffer, Context context, char ch)
{
  Context sc = space_context (context);
  if (context.tt)
    append (buffer, sc, ch);
  else if (was_whitespace (buffer, sc))
    ;
  else
    append (buffer, sc, ' ');
}


/* Parsing */

static void
syntax_error (const char *reason)
{
  error ("Bad input syntax: %s", reason);
}


static FILE *input;
static Buffer *buffer;
static int ch;			/* the next uneaten charcter */

typedef enum { B_tag, EM, I, PL, S, TT, U, DIGIT, COLOR } Tag_type;

enum { TAG, WHITESPACE, CHAR, END } token; /* the next uneaten token */
int      token_value;
Tag_type token_tag_type;
Flag     token_opening;


static void
report_token (void)
{
  char type = token == TAG ? (token_opening ? '+' : '-') : token == WHITESPACE ? 'w' : token == CHAR ? 'c' : 'e';

  fprintf (stderr, "%c %d %d\n", 
	   type, token_value, token_tag_type);
}


static void
nibble (void)
{
  ch = getc (input);
}


static void
gobble1 (int type)
{
  token_tag_type = type;
  token_value = 0;
  nibble ();
}


static void
gobble2 (int type, char expected)
{
  gobble1 (type);
  if (ch != expected)
    syntax_error ("Unexpected character");
  nibble ();
}


static void
eat_tag_name (void)
{
  switch (ch)
    {
    case 'B':      gobble1 (B_tag); break;
    case 'E':      gobble2 (EM, 'M'); break;
    case 'I':      gobble1 (I); break;
    case 'P':      gobble2 (PL, 'L'); break;
    case 'S':      gobble1 (S); break;
    case 'T':      gobble2 (TT, 'T'); break;
    case 'U':      gobble1 (U); break;

    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      token_tag_type = DIGIT;
      token_value = ch - '0';
      nibble ();
      break;

    case 'r':
    case 'g':
    case 'b':
    case 'c':
    case 'm':
    case 'y':
    case 'k':
    case 'w':
      token_tag_type = COLOR; 
      token_value = ch; 
      nibble (); 
      break;

    default:
      syntax_error ("Tag name expected");
      break;
    }
}


static void
eat (void)
{
  token = token_tag_type = token_value = token_opening = -1;

  switch (ch)
    {
    case '<': 
      token = TAG;
      nibble ();
      if (ch == '/')
	{
	  token_opening = false;
	  nibble ();
	}
      else
	token_opening = true;
      eat_tag_name ();
      if (ch != '>')
	syntax_error ("'>' expected");
      nibble ();
      break;

    case '\r':
    case '\n':
    case '\t':
    case ' ':
      token = WHITESPACE;
      token_value = ch;
      nibble ();
      break;

    case '>':
      syntax_error ("Unexpected '>'");
      break;

    case EOF:
      token = END;
      break;

    default:
      /* TODO: check here if it's printable? */
      token = CHAR;
      token_value = ch;
      nibble ();
      break;
    }

  if (0) report_token ();
}


static Context
new_context (Context context)
{
  Context result = context;

  switch (token_tag_type)
    {
    case B_tag:
      result.b = 1;
      break;

    case EM:
      if (!result.s)
	result.em = !result.em;
      break;

    case I:
      result.i = 1;
      break;

    case PL:
      result.s = result.em = result.i = result.b = result.tt = 0;
      result.u = 0;
      break;

    case S:
      result.s = 1;
      result.em = 0;
      break;

    case TT:
      result.tt = 1;
      break;

    case U:
      if (result.u < 3)
	result.u++;
      break;

    case DIGIT:
      result.size = token_value;
      break;

    case COLOR:
      switch (token_value)
	{
	case 'r': result.color = R; break;
	case 'g': result.color = G; break;
	case 'b': result.color = B_color; break;
	case 'c': result.color = C; break;
	case 'm': result.color = M; break;
	case 'y': result.color = Y; break;
	case 'k': result.color = K; break;
	case 'w': result.color = W; break;
	default: assert (false);
	}
      break;
	  
    default:
      printf ("%d\n", token_tag_type);
      assert (false);
    }

  check_context (result);
  return result;
}


static void
parse (Context context)
{
  for (;;)
    {
      switch (token)
	{
	case TAG:
	  if (!token_opening)
	    return;
	  {
	    Context subcontext = new_context (context);
	    Tag_type tt = token_tag_type;
	    int tv = token_value;
	    eat ();
	    parse (subcontext);
	    if (token != TAG || 
		token_tag_type != tt || token_value != tv ||
		token_opening)
	      syntax_error ("Mismatched closing tag");
	    eat ();
	    break;
	  }

	case WHITESPACE:
	  append_whitespace_char (buffer, context, token_value);
	  eat ();
	  break;

	case CHAR: 
	  append (buffer, context, token_value);
	  eat ();
	  break;

	case END:
	  return;
      
	default:
	  assert (false);
	  break;
	}
    }
}


static void
parse_document (FILE *doc_file, Buffer *the_buffer)
{
  input = doc_file;
  buffer = the_buffer;
  nibble ();
  eat ();
  parse (make_root_context ());
  assert (token == END);
}


/* Output */

static void
maybe_output_tag (const char *name, Flag tag, Flag opening)
{
  if (tag)
    {
      putc ('<', stdout);
      if (!opening)
	putc ('/', stdout);
      fputs (name, stdout);
      putc ('>', stdout);
    }
}


enum { bit_B, bit_EM, bit_I, bit_PL, bit_S, bit_TT, 
       bit_U1, bit_U2, bit_U3,
       bit_0, bit_1, bit_2, bit_3, bit_4, 
       bit_5, bit_6, bit_7, bit_8, bit_9, 
       bit_r, bit_g, bit_b, bit_c, bit_m, bit_y, bit_k, bit_w };

static const char *bit_to_tag_name[] = {
  "B", "EM", "I", "PL", "S", "TT", 
  "U", "U", "U", 
  "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", 
  "r", "g", "b", "c", "m", "y", "k", "w"
};

/*
typedef struct Node
{
  unsigned int bits;
  struct Node *kid, *sib;
}
Node;

static Node *
factor (Node *n)
{
  Node *sib = n->sib;
  if (sib)
    {
      unsigned common = n->bits & sib->bits;
      if (common != 0)
	{
	  Node *x = make_node (n);
	  ...
	}
    }
}
*/

static void
output_tags_from_bits (unsigned int bits, Flag opening)
{
  int i;
  if (opening)
    for (i = 0; i <= bit_w; ++i)
      maybe_output_tag (bit_to_tag_name[i], bits & (1 << i), opening);
  else
    for (i = bit_w; 0 <= i; --i)
      maybe_output_tag (bit_to_tag_name[i], bits & (1 << i), opening);
}


/* Return b1 & ~b2 except only on the high-order bits in common. */
static unsigned int
bits_diff (unsigned int b1, unsigned int b2)
{
  int i;
  unsigned int diff = b1 ^ b2;
  for (i = 0; i <= bit_w; ++i) 
    if (diff & (1 << i))
      break;
    else
      b1 = (b1 & ~(1 << i)) | ((b1 & ~b2) & (1 << i));
  return b1;
}


static unsigned int
tags_to_bits (Context context)
{
  Flag em = context.em && !context.s;

  int bits =
      (context.b << bit_B)
    | (em << bit_EM)
    | (context.i << bit_I)
    | (context.s << bit_S)
    | (context.tt << bit_TT)
    | (context.u >= 1 ? (1 << bit_U1) : 0)
    | (context.u >= 2 ? (1 << bit_U2) : 0)
    | (context.u == 3 ? (1 << bit_U3) : 0)
    | (context.size < Root_size ? (1 << (bit_0 + context.size)) : 0)
    | (context.color < Root_color ? (1 << (bit_r + context.color)) : 0)
    ;
    
  return bits;
}


static void
adjust_context (Context before, Context after)
{
  int before_bits = tags_to_bits (before);
  int after_bits = tags_to_bits (after);

  output_tags_from_bits (bits_diff (before_bits, after_bits), false);
  output_tags_from_bits (bits_diff (after_bits, before_bits), true);
}


static Context
bleed_forward (Context before, Context after)
{
  after.s = before.s;
  after.em = before.em;
  after.i = before.i;
  after.b = before.b;
  if (after.u == 0)
    after.color = before.color;
  return after;
}


/* Adjust context values of space characters for output correctness. */
static void
bleed (Buffer *buffer)
{
  int i;
  Context c = make_root_context ();

  for (i = 0; i < buffer->ptr; ++i)
    {
      Element *e = buffer->elements + i;
      if (isspace (e->character))
	e->context = bleed_forward (c, e->context);
      c = e->context;
    }
}


static void
output (const Buffer *buffer)
{
  int i;
  Context c = make_root_context ();

  for (i = 0; i < buffer->ptr; ++i)
    {
      const Element e = buffer->elements[i];
      adjust_context (c, e.context);
      putc (e.character, stdout);
      c = e.context;
    }

  adjust_context (c, make_root_context ());
}


/* Main program */

static Buffer *
load (const char *filename)
{
  FILE *fp = open_file (filename, "r");
  Buffer *result = make_buffer ();
  parse_document (fp, result);
  fclose (fp);
  return result;
}


int 
main (int argc, char **argv)
{
  Buffer *buffer = load (argv[1]);
  bleed (buffer);
  output (buffer);
  return 0;
}
